import os

from .authenticate import AuthCtx
from .printer import Printer
from .scanner import Scanner


class Client:
    EC_BASE_URL = 'https://api.epsonconnect.com'

    def __init__(self, base_url='', printer_email='', client_id='', client_secret='') -> None:
        base_url = base_url or self.EC_BASE_URL

        printer_email = printer_email or os.environ.get('ESPON_CONNECT_API_PRINTER_EMAIL')
        if not printer_email:
            raise ValueError('Printer Email can not be empty')

        client_id = client_id or os.environ.get('ESPON_CONNECT_API_CLIENT_ID')
        if not client_id:
            raise ValueError('Client ID can not be empty')

        client_secret = client_secret or os.environ.get('ESPON_CONNECT_API_CLIENT_SECRET')
        if not client_secret:
            raise ValueError('Client Secret can not be empty')

        self._auth_ctx = AuthCtx(base_url, printer_email, client_id, client_secret)

    def deauthticate(self):
        self._auth_ctx._deauthticate()

    @property
    def printer(self):
        return Printer(self._auth_ctx)

    @property
    def scanner(self):
        return Scanner(self._auth_ctx)



import pathlib
from urllib.parse import parse_qs, urlencode, urlparse

from .authenticate import AuthCtx
from .print_settings import (
    PrintSettingError,
    merge_with_default_settings,
    validate_settings,
)


class Printer:

    VALID_EXTENSIONS = {
        'doc',
        'docx',
        'xls',
        'xlsx',
        'ppt',
        'pptx',
        'pdf',
        'jpeg',
        'bmp',
        'gif',
        'png',
        'tiff',
    }

    def __init__(self, auth_ctx: AuthCtx) -> None:
        self._auth_ctx = auth_ctx

    @property
    def device_id(self):
        return self._auth_ctx.device_id

    def capabilities(self, mode):
        """
        Get device print capabilities.
        """
        method = 'GET'
        path = f'/api/1/printing/printers/{self.device_id}/capability/{mode}'

        return self._auth_ctx.send(method, path)

    def _print_setting(self, settings) -> dict:
        """
        Create a print job.
        """
        method = 'POST'
        path = f'/api/1/printing/printers/{self.device_id}/jobs'

        return self._auth_ctx.send(method, path, settings)

    def _upload_file(self, upload_uri: str, file_path: str, print_mode: str) -> None:
        """
        Upload file to be printed.
        """
        # Get extension from file path.
        extension = pathlib.Path(file_path).suffix.lower()
        if extension[1:] not in self.VALID_EXTENSIONS:
            raise PrintSettingError(f'{extension} is not a valid printing extension.')

        o = urlparse(upload_uri)
        q_dict = parse_qs(o.query)
        q_dict['File'] = [f'1{extension}']
        path = o._replace(query=urlencode(q_dict)).geturl()

        content_type = 'application/octet-stream'
        if print_mode == 'photo':
            content_type = 'image/jpeg'

        with open(file_path, 'rb') as fp:
            data = fp.read()

        headers = {
            'Content-Type': content_type,
            'Content-Length': str(len(data)),
        }

        method = 'POST'
        self._auth_ctx.send(method, path, data=data, headers=headers)

    def _execute_print(self, job_id):
        """
        Execute print job.
        """
        method = 'POST'
        path = f'/api/1/printing/printers/{self.device_id}/jobs/{job_id}/print'
        self._auth_ctx.send(method, path)

    def print(self, file_path, settings=None) -> str:
        """
        Print file.

        :return: Job ID for print job.
        """
        settings = merge_with_default_settings(settings)
        validate_settings(settings)

        job_data = self._print_setting(settings)
        self._upload_file(job_data['upload_uri'], file_path, settings['print_mode'])
        self._execute_print(job_data['id'])
        return job_data['id']

    def cancel_print(self, job_id):
        """
        Cancel print.
        """
        method = 'POST'
        path = f'/api/1/printing/printers/{self.device_id}/jobs/{job_id}/cancel'

    def job_info(self, job_id):
        """
        Get print job information.
        """
        method = 'GET'
        path = f'/api/1/printing/printers/{self.device_id}/jobs/{job_id}'

    def info(self):
        """
        Get device information.
        """
        method = 'GET'
        path = f'/api/1/printing/printers/{self.device_id}'


    def notifications(self):
        """
        Set notifications.
        """
        method = 'POST'
        path = f'/api/1/printing/printers/{self.device_id}/settings/notifications'



from .authenticate import AuthCtx


class Scanner:
    def __init__(self, auth_ctx: AuthCtx) -> None:
        self._auth_ctx = auth_ctx
        self._path = f'/api/1/scanning/scanners/{self._auth_ctx.device_id}/destinations'

    def list(self):
        """
        Get scan destinations.
        """
        method = 'GET'

    def add(self):
        """
        Register scan destination.
        """
        method = 'POST'

    def update(self):
        """
        Update scan destination.
        """
        method = 'POST'

    def remove(self):
        """
        Update scan destination.
        """
        method = 'DELETE'



from datetime import datetime, timedelta
import os

import requests
from requests.auth import HTTPBasicAuth


class AuthCtx:
    def __init__(
            self,
            base_url: str,
            printer_email: str,
            client_id: str,
            client_secret: str,
    ) -> None:
        self._base_url = base_url
        self._printer_email = printer_email
        self._client_id = client_id
        self._client_secret = client_secret

        self._expires_at = datetime.now()
        self._access_token = ''
        self._refresh_token = ''
        self._subject_id = ''

        self._auth()

    def _auth(self):
        method = 'POST'
        path = '/api/1/printing/oauth2/auth/token?subject=printer'

        if self._expires_at > datetime.now():
            return

        headers = {
            'Content-Type':'application/x-www-form-urlencoded'
        }
        auth = HTTPBasicAuth(self._client_id, self._client_secret)

        if self._access_token == '':
            data = {
                'grant_type': 'password',
                'username': self._printer_email,
                'password': '',
            }
        else:
            data = {
                'grant_type': 'refresh_token',
                'refresh_token': self._refresh_token,
            }

        body = self.send(method, path, data, headers=headers, auth=auth)

        error = body.get('error')
        if error:
            raise AuthenticationError(error)

        # First time authenticating, set refresh_token.
        if self._access_token == '':
            self._refresh_token = body['refresh_token']

        self._expires_at = datetime.now() + timedelta(seconds=int(body['expires_in']))
        self._access_token = body['access_token']
        self._subject_id = body['subject_id']

    def _deauthticate(self):
        """
        Cancel authentication.
        """
        method = 'DELETE'
        path = f'/api/1/printing/printers/{self._subject_id}'
        self.send(method, path)

    def send(self, method, path, data=None, headers=None, auth=None) -> dict:
        if not headers:
            headers = {
                'Authorization': f'Bearer {self._access_token}',
                'Content-Type': f'application/json',
            }

        resp = requests.request(
                method=method,
                url=self._base_url + path,
                headers=headers,
                data=data,
                auth=auth,
        ).json()

        error = resp.get('code')
        if error:
            raise ApiError(error)

        return resp

    @property
    def device_id(self):
        return self._subject_id


class AuthenticationError(RuntimeError):
    """
    Error for authentication specific exceptions.
    """


class ApiError(RuntimeError):
    """
    General base error for any API errors after authentication has succeeded.
    """


if __name__ == '__main__':
    # For testing.
    AuthCtx(
        base_url='https://api.epsonconnect.com',
        printer_email=os.environ['ESPON_CONNECT_API_PRINTER_EMAIL'],
        client_id=os.environ['ESPON_CONNECT_API_CLIENT_ID'],
        client_secret=os.environ['ESPON_CONNECT_API_CLIENT_SECRET'],
    )



import random
import string


VALID_PRINT_MODES = {
    'document',
    'photo',
}

VALID_MEDIA_SIZES = {
    'ms_a3',
    'ms_a4',
    'ms_a5',
    'ms_a6',
    'ms_b5',
    'ms_tabloid',
    'ms_letter',
    'ms_legal',
    'ms_halfletter',
    'ms_kg',
    'ms_l',
    'ms_2l',
    'ms_10x12',
    'ms_8x10',
    'ms_hivision',
    'ms_5x8',
    'ms_postcard',
}

VALID_MEDIA_TYPES = {
    'mt_plainpaper',
    'mt_photopaper',
    'mt_hagaki',
    'mt_hagakiphoto',
    'mt_hagakiinkjet',
}

VALID_PRINT_QUALITIES = {
    'high',
    'normal',
    'draft',
}

VALID_PAPER_SOURCES = {
    'auto',
    'rear',
    'front1',
    'front2',
    'front3',
    'front4',
}

VALID_COLOR_MODES = {
    'color',
    'mono',
}

VALID_TWO_SIDE = {
    'none',
    'long',
    'short',
}


def merge_with_default_settings(settings=None):
    settings = settings or {}

    job_name = settings.get('job_name') or ''
    if not job_name:
        # Generate random name if one is not given.
        job_name = ''.join(random.choice(string.ascii_letters) for _ in range(25))

    settings['job_name'] = job_name

    settings['print_mode'] = settings.get('print_mode') or 'document'

    print_setting = settings.get('print_setting') or {}

    # print_setting is optional so we can exit early if its not present.
    if not print_setting:
        return settings

    # Media Size
    print_setting['media_size'] = print_setting.get('media_size') or 'ms_a4'

    # media_type
    print_setting['media_type'] = print_setting.get('media_type') or 'mt_plainpaper'

    # borderless
    print_setting['borderless'] = print_setting.get('borderless') or False

    # print_quality
    print_setting['print_quality'] = print_setting.get('print_quality') or 'normal'

    # print_quality
    print_setting['source'] = print_setting.get('source') or 'auto'

    # color_mode
    print_setting['color_mode'] = print_setting.get('color_mode') or 'color'

    # two_sided
    print_setting['2_sided'] = print_setting.get('2_sided') or 'none'

    # reverse_order
    reverse_order = print_setting.get('reverse_order')
    if not reverse_order:
        reverse_order = False

    if print_setting['2_sided'] in ('long', 'short'):
        reverse_order = False

    print_setting['reverse_order'] = reverse_order

    # copies
    print_setting['copies'] = print_setting.get('copies') or 1

    # collate
    collate = print_setting.get('collate')
    if not collate:
        collate = True

    if print_setting['2_sided'] in ('long', 'short'):
        collate = True

    print_setting['collate'] = collate

    settings['print_setting'] = print_setting

    return settings


def validate_settings(settings: dict):
    """
    Validate all parts of a settings object.
    """
    extra_keys = set(settings.keys()) - {'job_name', 'print_mode', 'print_setting'}
    if extra_keys:
        raise PrintSettingError(f'Invalid settings keys {extra_keys}.')

    job_name = settings['job_name']
    if len(job_name) > 256:
        raise PrintSettingError(f'Job name is greater than 256 chars: {job_name}')

    print_mode = settings['print_mode']
    if print_mode not in VALID_PRINT_MODES:
        raise PrintSettingError(f'Not a valid print mode {print_mode}')

    # print_setting is optional so we can exit early if its not present.
    print_setting = settings.get('print_setting')
    if not print_setting:
        return

    # Media Size
    media_size = print_setting['media_size']
    if media_size not in VALID_MEDIA_SIZES:
        raise PrintSettingError(f'Invalid paper size {media_size}')

    # media_type
    media_type = print_setting['media_type']
    if media_type not in VALID_MEDIA_TYPES:
        raise PrintSettingError(f'Invalid media type {media_size}')

    # borderless
    borderless = print_setting['borderless']
    if not isinstance(borderless, bool):
        raise PrintSettingError(f'borderless must be a bool')

    # print_quality
    print_quality = print_setting['print_quality']
    if print_quality not in VALID_PRINT_QUALITIES:
        raise PrintSettingError(f'Invalid print quality {print_quality}')

    # print_quality
    source = print_setting['source']
    if source not in VALID_PAPER_SOURCES:
        raise PrintSettingError(f'Invalid source {source}')

    # color_mode
    color_mode = print_setting['color_mode']
    if color_mode not in VALID_COLOR_MODES:
        raise PrintSettingError(f'Invalid color mode {color_mode}')

    # two_sided
    two_sided = print_setting['2_sided']
    if two_sided not in VALID_TWO_SIDE:
        raise PrintSettingError(f'Invalid 2-sided value {two_sided}')

    # reverse_order
    reverse_order = print_setting['reverse_order']
    if not isinstance(reverse_order, bool):
        raise PrintSettingError(f'Reverse order must be a bool')

    # copies
    copies = print_setting['copies']
    if copies < 1 or copies > 99:
        raise PrintSettingError(f'Invalid number of copies {copies}')

    # collate
    collate = print_setting['collate']
    if not isinstance(collate, bool):
        raise PrintSettingError(f'Collate must be a bool')


class PrintSettingError(ValueError):
    pass
